#ifndef LIST_HPP
#define LIST_HPP

#include <iostream>

// Előzetes deklarációk a barát-operátor kereszthivatkozása miatt.
template <typename T>
class List;

template <typename T>
std::ostream &operator<<(std::ostream &out, const List<T> &l);


// Egyirányú, fejelemes láncolt lista sablon.
template <typename T>
class List
{
	// Publikus rész (implementáció).
	public:
		// Kivételek.
		enum ListException
		{
			// Azt az esetet jelöli, amikor a lista üres, de az adott művelet így nem értelmezhető.
			Empty
		};

		// Konstruktor: létrehozza a láncolt lista fejelemét és beállítja a mutatókat.
		List();
		// Copy konstruktor: lemásolja az argumentumban megadott objektumot.
		List(const List &l);
		// Destruktor, a HEAP-en lefoglalt terület felszabadítása.
		virtual ~List();
		// Értékadás operátor: lemásolja az argumentumban megadott objektumot, ha nem saját magát adtuk értékül.
		List &operator=(const List &ol);
		
		// Elem hozzáfűzése a láncolt lista végére.
		void append(const T &v);
		// Elem befűzése a láncolt lista elejére.
		void prepend(const T &v);
		// A láncolt listában tárolt maximális érték megkeresése. Ha a lista üres, akkor kivétel kiváltása.
		T getMax() const throw(ListException);
		
		// Kiíró operátor megvalósítása, a láncolt lista elemeinek felsorolása.
		friend std::ostream &operator<<<T>(std::ostream &out, const List &l);
		
	// Privát rész (reprezentáció).
	private:
		// A láncolt lista elemeinek reprezentációja.
		struct Node
		{
			// Tárolt érték.
			T value;
			// Következő elem mutatója.
			Node* next;
			
			// Konstruktor: létrehoz egy új láncolt lista elemet az alapértelmezett értékkel.
			Node(Node *n = NULL) : next(n) {}
			// Konstruktor: létrehoz egy új láncolt lista elemet a megadott értékkel.
			Node(const T &v, Node *n = NULL) : value(v), next(n) {}
		};
		
		// Fejmutató és az utolsó elem mutatója.
		Node *header, *last;
};

// Konstruktor: létrehozza a láncolt lista fejelemét és beállítja a mutatókat.
template <typename T>
List<T>::List()
{
	header = new Node();
	last = header;
}

// Copy konstruktor: lemásolja az argumentumban megadott objektumot.
template <typename T>
List<T>::List(const List &ol)
{
	header = new Node(ol.header->value);
	last = header;
	
	Node *current = ol.header->next;
	while(current != NULL)
	{
		append(current->value);
		current = current->next;
	}
}

// Destruktor, a HEAP-en lefoglalt terület felszabadítása.
template <typename T>
List<T>::~List()
{
	Node *current;
	while(header != NULL)
	{
		current = header;
		header = header->next;
		delete current;
	}
}

// Értékadás operátor: lemásolja az argumentumban megadott objektumot, ha nem saját magát adtuk értékül.
template <typename T>
List<T> &List<T>::operator=(const List<T> &ol)
{
	// Saját magának próbáljuk-e értékül adni?
	if(this == &ol)
		return *this;

	// Régi lista megsemmisítése.
	Node *current;
	while(header != NULL)
	{
		current = header;
		header = header->next;
		delete current;
	}

	// Paraméter lista átmásolása.
	header = new Node(ol.header->value);
	last = header;
	
	current = ol.header->next;
	while(current != NULL)
	{
		append(current->value);
		current = current->next;
	}
	
	return *this;
}

// Elem hozzáfűzése a láncolt lista végére.
template <typename T>
void List<T>::append(const T &v)
{
	last->next = new Node(v);
	last = last->next;
}

// Elem befűzése a láncolt lista elejére.
template <typename T>
void List<T>::prepend(const T &v)
{
	header->next = new Node(v, header->next);
}

// A láncolt listában tárolt maximális érték megkeresése. Ha a lista üres, akkor kivétel kiváltása.
template <typename T>
T List<T>::getMax() const throw(ListException)
{
	if(header->next == NULL)
		throw Empty;
	
	Node *max = header->next;
	Node *current = max->next;
	while(current != NULL)
	{
		if(current->value > max->value)
			max = current;
		current = current->next;
	}
	return max->value;
}

// Kiíró operátor megvalósítása, a láncolt lista elemeinek felsorolása.
template <typename T>
std::ostream &operator<<(std::ostream &out, const List<T> &l)
{
	// Fontos a typename kulcsszó, hogy a Node-ot típusként értelmezze a fordító!
	typename List<T>::Node *current = l.header->next;
	while(current != NULL)
	{
		out << current->value << " ";
		current = current->next;
	}
	out << std::endl;
}

#endif
