#ifndef MATRIX_HHPP
#define MATRIX_HHPP

#include <iostream>

// Előzetes deklarációk a barát-operátor kereszthivatkozása miatt.
template <typename T = int>
class Matrix;

template <typename T>
std::ostream& operator<<(std::ostream& out, const Matrix<T>& m);

// Mátrix sablon osztály.
template <typename T>
class Matrix
{
	// Publikus rész (implementáció).
	public:
		// Kivételek.
		enum Exception
		{
			// Azt az esetet jelöli, ha a megadott index kívül esett a mátrixon.
			OutOfBoundary,
			// Azt az esetet jelöli, ha a paraméter mátrixok mérete nem egyezik, pedig ez a művelethez szükséges lenne.
			SizeNotMatch
		};
		
		// Konstruktor: beállítja a mátrix méretét és allokálja a szükséges területet a HEAP-en.
		Matrix(int n, int m);
		// Másoló (Copy) konstruktor: lemásolja az argumentumban megadott objektumot.
		Matrix(const Matrix &om);
		// Destruktor, a HEAP-en lefoglalt terület felszabadítása.
		virtual ~Matrix();
		// Értékadás operátor: lemásolja az argumentumban megadott objektumot, ha nem saját magát adtuk értékül.
		Matrix &operator=(const Matrix &om);
	
		// Mátrix méretének lekérdezése.
		T N() const { return n; }
		T M() const { return m; }
		
		// A mátrix feltöltése egy adott értékkel.
		void fill(T v);
		// Egy adott pozíció értékét meghatározza. Ha a pozíció a mátrixon kívül esik, akkor kivétel kiváltása.
		void insert(int i, int j, T v) throw(Exception);
		// Egy adott pozíció értékének visszaadása. Ha a pozíció a mátrixon kívül esik, akkor kivétel kiváltása.
		T get(int i, int j) const throw(Exception);
		
		// Összeadás operátor: összeadja a jelen és a paraméter mátrixot és egy harmadik objektumként visszaadja.
		// Ha a két mátrix mérete nem egyezik, akkor kivétel kiváltása.
		Matrix operator+(const Matrix &om) const throw(Exception);
		// Zárójel operátor egy adott pozíció értékének visszaadására.
		// Ha a pozíció a mátrixon kívül esik, akkor kivétel kiváltása.
		T operator()(int i, int j) const throw(Exception);
		// Zárójel operátor egy adott pozíció értékének referencia szerinti visszaadására.
		// Ha a pozíció a mátrixon kívül esik, akkor kivétel kiváltása.
		T& operator()(int i, int j) throw(Exception);
		
		// A mátrix kiírása adatfolyamra.
		friend std::ostream& operator<<<T>(std::ostream& out, const Matrix& m);
	
	// Privát rész (reprezentáció).
	private:
		// A mátrix mérete.
		int n, m;
		// A mátrix elemei.
		T** matrix;
};

// Konstruktor: beállítja a mátrix méretét és allokálja a szükséges területet a HEAP-en.
template <typename T>
Matrix<T>::Matrix(int n, int m)
{
	this->n = n;
	this->m = m;
	
	matrix = new T*[n];
	for(int i = 0; i < n; ++i)
		matrix[i] = new T[m];
}

// Másoló (Copy) konstruktor: lemásolja az argumentumban megadott objektumot.
template <typename T>
Matrix<T>::Matrix(const Matrix &om)
{
	n = om.n;
	m = om.m;
	
	matrix = new T*[n];
	for(int i = 0; i < n; ++i)
	{
		matrix[i] = new T[m];
		for(int j = 0; j < m; ++j)
			matrix[i][j] = om.matrix[i][j];
	}
}

// Destruktor, a HEAP-en lefoglalt terület felszabadítása.
template <typename T>
Matrix<T>::~Matrix()
{
	for(int i = 0; i < n; ++i)
		delete[] matrix[i];
	delete[] matrix;
}

// Értékadás operátor: lemásolja az argumentumban megadott objektumot, ha nem saját magát adtuk értékül.
template <typename T>
Matrix<T> &Matrix<T>::operator=(const Matrix &om)
{
	if(this == &om)
		return *this;

	for(int i = 0; i < n; ++i)
		delete[] matrix[i];
	delete[] matrix;

	n = om.n;
	m = om.m;
	
	matrix = new T*[n];
	for(int i = 0; i < n; ++i)
	{
		matrix[i] = new T[m];
		for(int j = 0; j < m; ++j)
			matrix[i][j] = om.matrix[i][j];
	}
	
	return *this;
}

// A mátrix feltöltése egy adott értékkel.
template <typename T>
void Matrix<T>::fill(T v)
{
	for(int i = 0; i < n; ++i)
		for(int j = 0; j < m; ++j)
			matrix[i][j] = v;
}

// Egy adott pozíció értékét meghatározza. Ha a pozíció a mátrixon kívül esik, akkor kivétel kiváltása.
template <typename T>
void Matrix<T>::insert(int i, int j, T v) throw(Exception)
{
	if(i >= 0 && i < n && j >= 0 && j < m)
		matrix[i][j] = v;
	else
		throw OutOfBoundary;
}

// Egy adott pozíció értékének visszaadása. Ha a pozíció a mátrixon kívül esik, akkor kivétel kiváltása.
template <typename T>
T Matrix<T>::get(int i, int j) const throw(Exception)
{
	if(i >= 0 && i < n && j >= 0 && j < m)
		return matrix[i][j];
	else
		throw OutOfBoundary;
}

// Összeadás operátor: összeadja a jelen és a paraméter mátrixot és egy harmadik objektumként visszaadja.
// Ha a két mátrix mérete nem egyezik, akkor kivétel kiváltása.
template <typename T>
Matrix<T> Matrix<T>::operator+(const Matrix &om) const throw(Exception)
{
	if(n == om.n && m == om.m)
	{
		Matrix r(n, m);
		for(int i = 0; i < n; ++i)
			for(int j = 0; j < m; ++j)
				r.insert(i, j, matrix[i][j] + om.matrix[i][j]);
		return r;
	}
	else
		throw SizeNotMatch;
	
}

// Zárójel operátor egy adott pozíció értékének visszaadására.
// Ha a pozíció a mátrixon kívül esik, akkor kivétel kiváltása.
template <typename T>
T Matrix<T>::operator()(int i, int j) const throw(Exception)
{
	return get(i, j);
}

// Zárójel operátor egy adott pozíció értékének referencia szerinti visszaadására.
// Ha a pozíció a mátrixon kívül esik, akkor kivétel kiváltása.
template <typename T>
T& Matrix<T>::operator()(int i, int j) throw(Exception)
{
	if(i >= 0 && i < n && j >= 0 && j < m)
		return matrix[i][j];
	else
		throw OutOfBoundary;
}

// A mátrix kiírása adatfolyamra.
template <typename T>
std::ostream& operator<<(std::ostream& out, const Matrix<T>& m)
{
	for(int i = 0; i < m.n; ++i)
	{
		for(int j = 0; j < m.m; ++j)
			out << m.matrix[i][j] << " ";
		out << std::endl;
	}
	return out;
}

#endif
